/*
  Originally obtained from Phil Church - pchurch@lanl.gov
  muse.inst.c
  Dynamically records the low and high watermarks for heap and stack
  and calculates the maximum heap and stack size for the program.
  Stack limits are obtained from the Stack Pointer (SP) while heap
  limits are calculated by intercepting calls to brk(), mmap(), sbrk()
*/

#include <stdio.h>
#include <string.h>
#include <cmplrs/atom.inst.h>

static int check_obj_name( Obj * );
static const char *get_name( Obj * );

/* perform once-per-application initializations */
void InstrumentInit( int iargc, char **iargv )
{
  unsigned long  BrkStart;
  Obj * ObjMain;
  /* define prototypes for the analysis routines */
  AddCallProto("Initialize(long, REGV)");
  AddCallProto("BeforeBrk(REGV)");
  AddCallProto("BeforeSbrk(REGV)");
  AddCallProto("BeforeMmap(REGV)");
  AddCallProto("PrintResults()");
  AddCallProto("GetStack(REGV)");
  /* Calculate the start of this apps heap and pass it to the
     initialization routine in the analysis code.  The application's
     heap starts immediately after the main object's uninitialized
     data area.  Also pass the SP as the base of the stack area. */
  ObjMain = GetFirstObj();
  BrkStart = GetObjInfo(ObjMain, ObjUninitDataStartAddress) +
    GetObjInfo(ObjMain, ObjUninitDataSize);
  AddCallProgram(ProgramBefore, "Initialize", BrkStart, REG_SP );
  /* print out the results when app finishes */
  AddCallProgram(ProgramAfter, "PrintResults" );
}
/* Instrument each object in this app */
Instrument( int iargc, char **iargv, Obj *obj )
{
  ProcRes pres;
  Proc *p;
  /* capture the Stack pointer at each procedure call
     This allows us to keep track of the end of the apps stack */
  for( p = GetFirstObjProc(obj); p != NULL; p = GetNextProc(p))
    {
      AddCallProc(p, ProcAfter, "GetStack", REG_SP );
    }
  /* If this object defines brk(), add an instrumentation point.
     This allows us to keep track of the end of the apps heap */
  ResolveNamedProc("__brk", &pres);
  if( pres.proc != NULL )
    {
      if( check_obj_name(pres.obj) )
        {
          printf( " defined brk()\n" );
          AddCallProc(pres.proc, ProcBefore, "BeforeBrk", REG_ARG_1 );
        }
    }
  /* If this object defines sbrk(), add an instrumentation point.
     This allows us to keep track of the end of the apps heap */
  ResolveNamedProc("__sbrk", &pres);
  if( pres.proc != NULL )
    {
      if( check_obj_name(pres.obj) )
        {
          printf( " defined sbrk()\n" );
          AddCallProc(pres.proc, ProcBefore, "BeforeSbrk", REG_ARG_1 );
        }
    }
  /* If this object defines mmap(), add an instrumentation point.
     This allows us to keep track of the end of the apps heap */
  ResolveNamedProc("__mmap", &pres);
  if( pres.proc != NULL )
    {
      if( check_obj_name(pres.obj) )
        {
          printf( " defined mmap()\n" );
          AddCallProc(pres.proc, ProcBefore, "BeforeMmap", REG_ARG_2 );
        }
    }
}
/* Ensure that the procedure being replaced is not a user written
   routine, but the expected standard function.  This test can only
   be done for shared libraries */
static int check_obj_name( Obj *obj )
{
  long otype;
  int result=0;
  if( obj )
    {
      otype = GetObjInfo( obj, ObjShared );
      /* if the object is sharable (.so), see if the routine is in
         the lib we expect it to be */
      if( otype == OBJ_SHARABLE )
        {
          if( strcmp( get_name(obj), "libc.so" ) == 0 )
            {
              result = 1;
            }
        }
      else if( otype == OBJ_NON_SHARED )
        {
          result=1;
        }
    }
  return( result );
}
/* get the simple name of an object.  Don't use basename(3) because its API
   requires ability to modify the string, and GetObjName return a const */
static const char *get_name( Obj *object )
{
  const char *pname = GetObjName( object );
  const char *pslash;
  /* Return pointer to char following last '/', or NULL if no name */
  if( !pname )
    {
      return NULL;
    }
  else
    {
      pslash = strrchr( pname, '/' );
    }
  if( pslash )
    {
      pname = ++pslash;
    }
  if( *pname == '\0' )
    {
      return NULL;
    }
  else
    {
      return( pname );
    }
}
