#include "procmon_info.h"
#include "procmon_rld.h"
/* ========================================================================

   Name
   ====
   PROCMON_rld_init

   Purpose
   =======
   Initialize data before sampling starts

   Arguments
   =========
   ierr          Intent: out
                 C type: (na) error return value
                 error return value

   Return Values
   =============
   0 if ok

   Program Flow
   ============
   1) process environment variables and set variables

   ======================================================================== */
int PROCMON_rld_init(
                     void
                     )
{
  int
    ierr=0, /* error return value */
    signal; /* which signal to use */
  long long
    part_sec, /* part of a second (some units like microseconds) */
    secs; /* number of seconds */
  /* ---------- */
  /* init pinfo */
  /* ---------- */
  /* --------------------------------------------------------------------- */
  /* do not initialize procmon_s.psettings.tag since this routine might    */
  /* get called (due to fork and needing to get new file) and previous tag */
  /* setting needs to still be there.                                      */
  /* --------------------------------------------------------------------- */
  PROCMON_init_file( &(procmon_s.pinfo), &(procmon_s.psettings),
                     &(procmon_s.filename_fp) );
  /* ------------------ */
  /* signal/timer stuff */
  /* ------------------ */
  procmon_s.chain_signal = 0; /* do not chain signal handlers */
  signal = procmon_s.psettings.signal;
#if defined(_PROCMON_ITIMER)
  /* ------------ */
  /* itimer stuff */
  /* ------------ */
  /* other pairs
     SIGVTALRM -> ITIMER_VIRTUAL
     SIGPROF   -> ITIMER_PROF
  */
  procmon_s.which_itimer              = ITIMER_REAL;
  part_sec = (procmon_s.psettings.time%1000)*1000;
  secs = procmon_s.psettings.time/1000;
  procmon_s.value.it_interval.tv_sec  = secs;
  procmon_s.value.it_interval.tv_usec = part_sec;
  procmon_s.value.it_value.tv_sec     = secs;
  procmon_s.value.it_value.tv_usec    = part_sec;
#else
  /* ------------------ */
  /* POSIX4 timer stuff */
  /* ------------------ */
  procmon_s.psxsigevent.sigev_notify          = SIGEV_SIGNAL;
  procmon_s.psxsigevent.sigev_signo           = signal;
  procmon_s.psxsigevent.sigev_value.sival_int = 13;
  part_sec = (procmon_s.psettings.time%1000)*1000000;
  secs     = procmon_s.psettings.time/1000;
  procmon_s.psxspec.it_interval.tv_nsec = part_sec;
  procmon_s.psxspec.it_interval.tv_sec  = secs;
  procmon_s.psxspec.it_value.tv_nsec    = part_sec;
  procmon_s.psxspec.it_value.tv_sec     = secs;
#endif
  /* --------------- */
  /* sigaction stuff */
  /* --------------- */
  procmon_s.which_sigaction = signal;
  memset(&(procmon_s.action_old),0x00,sizeof(struct sigaction));
  memset(&(procmon_s.action),0x00,sizeof(struct sigaction));
  procmon_s.action.sa_flags = SA_RESTART|SA_SIGINFO;
  procmon_s.action.sa_sigaction = PROCMON_rld_handler;
  memset(&(procmon_s.action_ignore),0x00,sizeof(struct sigaction));
  procmon_s.action_ignore.sa_handler = SIG_IGN;
  /* other possibility to avoid signal backup: change new handler to
     a no-op function.  For some reason, doesn't work.
    procmon_s.action_ignore.sa_flags = SA_RESTART|SA_SIGINFO;
    procmon_s.action_ignore.sa_sigaction = PROCMON_rld_handler_noop;
  */
  /* ------------------- */
  /* set output filename */
  /* ------------------- */
  /* ---------------------------- */
  /* open file and print a header */
  /* ---------------------------- */
  /* ---------------------------------------- */
  /* no need to mkstemp...just overwrite file */
  /* ---------------------------------------- */
  /*
    procmon_s.filename_fd = mkstemp( procmon_s.filename );
    if( procmon_s.filename_fd == -1 )
    abort();
    if( close( procmon_s.filename_fd ) != 0 )
    abort();
  */
  procmon_s.initialized = !0;
  return( ierr );
}
/* ========================================================================

   Name
   ====
   PROCMON_rld_terminate

   Purpose
   =======
   free stuff gotten in init call

   Arguments
   =========
   ierr          Intent: out
                 C type: (na) error return value
                 error return value

   Return Values
   =============
   0 if ok

   Program Flow
   ============
   1) process environment variables and set variables

   ======================================================================== */
int PROCMON_rld_terminate(
                          void
                          )
{
  int
    ierr=0; /* error return value */
  if( procmon_s.initialized == 0 )
    return( ierr );
  procmon_s.initialized = 0;
  if( procmon_s.filename_fp != NULL )
    {
      if( procmon_s.filename_fp != stdout )
        {
          ierr = fclose( procmon_s.filename_fp );
          assert( !ierr );
          /* --------------------------------- */
          /* delete file if not enough samples */
          /* --------------------------------- */
          if( procmon_s.pinfo.num_samples < procmon_s.psettings.min ||
              procmon_s.pinfo.num_samples_printed < 1 )
            {
              ierr = remove( procmon_s.psettings.outfile_name );
              assert( !ierr );
            }
        }
      procmon_s.filename_fp = NULL;
    }
  return( ierr );
}
/* ========================================================================

   Name
   ====
   PROCMON_rld_ctor

   Purpose
   =======
   constructor at process start

   Arguments
   =========
   void          Intent: in
                 C type: void
                 no arguments or return values

   Return Values
   =============
   none

   Program Flow
   ============
   1) init: PROCMON_rld_init
   2) get sampling at beginning
   3) set catcher first then start itimer in case timer goes off

   ======================================================================== */
void PROCMON_rld_ctor(
                      void
                      )
{
  /* --------- */
  /* init info */
  /* --------- */
  if( PROCMON_rld_init() != 0 )
    abort();
  /* ------------------------- */
  /* get sampling at beginning */
  /* ------------------------- */
  PROCMON_rld_handler(0,NULL,NULL);
  /* --------------------------------------------------------- */
  /* set catcher first then start timer in case timer goes off */
  /* --------------------------------------------------------- */
  if (sigaction( procmon_s.which_sigaction, &(procmon_s.action),
                &(procmon_s.action_old)) != 0)
    abort();
  PROCMON_rld_init_timer();
}
/* ========================================================================

   Name
   ====
   PROCMON_rld_dtor

   Purpose
   =======
   destructor at process termination

   Arguments
   =========
   void          Intent: in
                 C type: void
                 no arguments or return values

   Return Values
   =============
   none

   Program Flow
   ============
   1) reset old catcher after timer in case timer goes off
   2) get sampling at end
   3) free info

   ======================================================================== */
void PROCMON_rld_dtor(
                      void
                      )
{
  /* ---------------------------------------------------- */
  /* reset old catcher after timer in case timer goes off */
  /* ---------------------------------------------------- */
#if defined(_PROCMON_ITIMER)
  /* ------ */
  /* itimer */
  /* ------ */
  if (setitimer(procmon_s.which_itimer, &(procmon_s.value_old),
                &(procmon_s.value)) != 0)
    abort();
#else
  /* ------------ */
  /* POSIX4 timer */
  /* ------------ */
  if( timer_delete( procmon_s.psxtimer ) != 0 )
    abort();
#endif
  if (sigaction(procmon_s.which_sigaction, &(procmon_s.action_old),
                &(procmon_s.action)) != 0)
    abort();
  /* ------------------- */
  /* get sampling at end */
  /* ------------------- */
  PROCMON_rld_handler(0,NULL,NULL);
  /* --------- */
  /* free info */
  /* --------- */
  if( PROCMON_rld_terminate() != 0 )
    abort();
}
